import os
from contextlib import suppress
from typing import Optional

import pytest

from aiocouch import CouchDB

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_server_version(couchdb: CouchDB) -> None:
    response = await couchdb.info()

    from packaging import version

    assert version.parse("2.0.0") < version.parse(response["version"])


async def test_session(couchdb: CouchDB) -> None:
    headers, response = await couchdb._server._get("/_session")

    assert not isinstance(response, bytes)

    assert "ok" in response
    assert "userCtx" in response
    assert "name" in response["userCtx"]

    user: Optional[str] = None

    with suppress(KeyError):
        user = os.environ["COUCHDB_USER"]

    assert user == response["userCtx"]["name"]


async def test_cookie_authentication(couchdb_with_user_access: CouchDB) -> None:
    import os

    import aiohttp

    from aiocouch import CouchDB

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    user = "aiocouch_test_user"

    # create a session cookie, which can be used later
    async with aiohttp.ClientSession() as session:
        async with session.post(
            f"{hostname}/_session",
            data={"name": user, "password": user},
        ) as resp:
            assert resp.status == 200
            await resp.json()
            cookie = resp.cookies["AuthSession"].value

    async with CouchDB(hostname, cookie=cookie) as couchdb:
        await couchdb.check_credentials()

        headers, response = await couchdb._server._get("/_session")
        assert not isinstance(response, bytes)
        assert user == response["userCtx"]["name"]


async def test_basic_authentication() -> None:
    import os

    from aiocouch import CouchDB

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    user: Optional[str] = None
    with suppress(KeyError):
        user = os.environ["COUCHDB_USER"]

    try:
        password = os.environ["COUCHDB_PASS"]
    except KeyError:
        password = ""

    async with CouchDB(hostname, user=user, password=password) as couchdb:
        await couchdb.check_credentials()


async def test_with_wrong_credentials() -> None:
    import os

    from aiocouch import CouchDB, UnauthorizedError

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    with pytest.raises(UnauthorizedError):
        async with CouchDB(
            hostname,
            user="invalid",
            password="rcvy438tyb7est0fb38s4tybf74etbc7843tybfs4fb7v49bstf68bs495ftb63948ft5b9s6",
        ) as couchdb:
            await couchdb["does_not_exist"]


async def test_check_wrong_credentials() -> None:
    import os

    from aiocouch import CouchDB, UnauthorizedError

    try:
        hostname = os.environ["COUCHDB_HOST"]
    except KeyError:
        hostname = "http://localhost:5984"

    with pytest.raises(UnauthorizedError):
        async with CouchDB(
            hostname,
            user="invalid",
            password="rcvy438tyb7est0fb38s4tybf74etbc7843tybfs4fb7v49bstf68bs495ftb63948ft5b9s6",
        ) as couchdb:
            await couchdb.check_credentials()



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from contextlib import suppress
from typing import (
    Any,
    AsyncContextManager,
    AsyncGenerator,
    Callable,
    List,
    Optional,
    TypeVar,
)

from . import couchdb
from .bulk import BulkCreateOperation, BulkUpdateOperation
from .design_document import DesignDocument
from .document import Document, SecurityDocument
from .event import BaseChangeEvent, ChangedEvent, DeletedEvent
from .exception import ConflictError, NotFoundError
from .remote import RemoteDatabase
from .request import FindRequest
from .typing import JsonDict
from .view import AllDocsView, View

FuncT = TypeVar("FuncT", bound=Callable[..., Any])


def _returns_async_context_manager(f: FuncT) -> FuncT:
    setattr(f, "__returns_acontextmanager__", True)
    return f


class Database(RemoteDatabase):
    """A local representation for the referenced CouchDB database

    An instance of this class represents a local copy of a CouchDB database. It allows
    to create and retrieve :class:`~aiocouch.document.Document` instances, as well as
    the iteration other many documents.

    :ivar id: the id of the database

    :param `~aiocouch.CouchDB` couchdb: The CouchDB connection session
    :param id: the id of the database

    """

    def __init__(self, couchdb: "couchdb.CouchDB", id: str):
        super().__init__(couchdb._server, id)

    async def akeys(self, **params: Any) -> AsyncGenerator[str, None]:
        """A generator returning the names of all documents in the database

        :param params: passed into :meth:`aiohttp.ClientSession.request`
        :return: returns all document ids

        """
        async for key in self.all_docs.ids(**params):
            yield key

    async def create(
        self, id: str, exists_ok: bool = False, data: Optional[JsonDict] = None
    ) -> "Document":
        """Returns a local representation of a new document in the database

        This method will check if a document with the given name already exists and
        return a :class:`~aiocouch.document.Document` instance.

        This method does not create a document with the given name on the server. You
        need to call :meth:`~aiocouch.document.Document.save` on the returned document.

        :raises ~aiocouch.ConflictError: if the document does not exist on the server

        :param id: the name of the document
        :param exists_ok: If `True`, do not raise a :class:`~aiocouch.ConflictError` if
            an document with the given name already exists. Instead return the existing
            document.
        :param data: Description of parameter `data`. Defaults to None.
        :return: returns a :class:`~aiocouch.document.Document` instance representing
            the requested document

        """
        doc = Document(self, id, data=data)

        if exists_ok:
            with suppress(NotFoundError):
                await doc.fetch(discard_changes=True)
        else:
            if await doc._exists():
                raise ConflictError(
                    f"The document '{id}' does already exist in the database '{self.id}'"
                )

        return doc

    async def delete(self) -> None:
        """Delete the database on the server

        Send the request to delete the database and all of its documents.

        """
        await self._delete()

    async def docs(
        self,
        ids: Optional[List[str]] = None,
        create: bool = False,
        prefix: Optional[str] = None,
        include_ddocs: bool = False,
        **params: Any,
    ) -> AsyncGenerator["Document", None]:
        """A generator to iterator over all or a subset of documents in the database.

        When neither of ``ids`` nor ``prefix`` are specified, all documents will be
        iterated. Only one of ``ids`` and ``prefix`` can be specified. By default, design
        documents are not included.

        :param ids: Allows to iterate over a subset of documents by passing a list of
            document ids
        :param create: If ``True``, every document contained in `ids`, which doesn't
            exist, will be represented by an empty
            :class:`~aiocouch.document.Document` instance.
        :param prefix: Allows to iterator over a subset of documents by specifing a
            prefix that the documents must match.
        :param include_ddocs: Include the design documents of the database.
        :param params: Additional query parameters,
            see :ref:`CouchDB view endpoint <couchdb:api/ddoc/view>`.

        """
        if ids is not None and len(ids) == 0:
            return

        async for doc in self.all_docs.docs(
            ids, create, prefix, include_ddocs, **params
        ):
            yield doc

    async def values(self, **params: Any) -> AsyncGenerator["Document", None]:
        """Iterates over documents in the database

        See :meth:`~aiocouch.database.Database.docs`.
        """
        async for doc in self.all_docs.docs(**params):
            yield doc

    @property
    def all_docs(self) -> AllDocsView:
        """Returns the all_docs view of the database

        :return: Description of returned object.

        """
        return AllDocsView(self)

    def view(self, design_doc: str, view: str) -> View:
        return View(self, design_doc, view)

    async def design_doc(self, id: str, exists_ok: bool = False) -> DesignDocument:
        ddoc = DesignDocument(self, id)

        if exists_ok:
            with suppress(NotFoundError):
                await ddoc.fetch(discard_changes=True)
        else:
            if await ddoc._exists():
                raise ConflictError(
                    f"The design document '{id}' does already exist in the database '{self.id}'"
                )

        return ddoc

    async def find(
        self, selector: Any, limit: Optional[int] = None, **params: Any
    ) -> AsyncGenerator["Document", None]:
        """Fetch documents based on search criteria

        This method allows to use the :ref:`_find<couchdb:api/db/_find>`
        endpoint of the database.

        This method supports all request paramters listed in
        :ref:`_find<couchdb:api/db/_find>`.

        .. note:: As this method returns :class:`~aiocouch.document.Document` s, which
            contain the complete data, the `fields` parameter is not supported.

        :param type selector: See :ref:`selectors<couchdb:find/selectors>`
        :return: return all documents matching the passed selector.
        """

        # we need to get the complete doc, so fields selector isn't allowed
        if "fields" in params.keys():
            raise ValueError("The fields parameter isn't supported")

        async for doc in FindRequest(self, selector, limit, **params):
            yield doc

    async def index(self, index: JsonDict, **kwargs: Any) -> JsonDict:
        """Create a new index on the database

        This method allows to use the :ref:`_index<couchdb:api/db/find/index>`
        endpoint of the database.

        This method supports all request paramters listed in
        :ref:`_index<couchdb:api/db/find/index>`.

        :param index: JSON description of the index
        :param kwargs: additional parameters, refer to the CouchDB documentation
        :return: The response of the CouchDB _index endpoint
        """
        return await self._index(index, **kwargs)

    @_returns_async_context_manager
    def update_docs(
        self, ids: List[str] = [], create: bool = False
    ) -> AsyncContextManager[BulkUpdateOperation]:
        """Update documents in bulk.

        See :ref:`bulk operations`.

        :param ids: list of affected documents, defaults to []
        :param create: [description], defaults to False
        :return: A context manager for the bulk operation

        """

        return BulkUpdateOperation(self, ids, create)

    @_returns_async_context_manager
    def create_docs(
        self, ids: List[str] = []
    ) -> AsyncContextManager[BulkCreateOperation]:
        """Create documents in bulk.

        See :ref:`bulk operations`.

        :param ids: list of document ids to be created
        :return: A context manager for the bulk operation

        """

        return BulkCreateOperation(self, ids)

    async def __getitem__(self, id: str) -> Document:
        """Returns the document with the given id

        :raises `~aiocouch.NotFoundError`: if the given document does not exist

        :param id: the name of the document
        :return: a local copy of the document

        """
        return await self.get(id)

    async def get(
        self, id: str, default: Optional[JsonDict] = None, *, rev: Optional[str] = None
    ) -> Document:
        """Returns the document with the given id

        :raises `~aiocouch.NotFoundError`: if the given document does not exist and
            `default` is `None`
        :raises `~aiocouch.BadRequestError`: if the given rev of the document is
            invalid or missing


        :param id: the name of the document
        :param default: if `default` is not `None` and the document does not exist on
            the server, a new :class:`~aiocouch.document.Document` instance, containing
            `default` as its contents, is returned. To create the document on the
            server, :meth:`~aiocouch.document.Document.save` has to be called on the
            returned instance.
        :param rev: The requested rev of the document. The requested rev might not
            or not anymore exist on the connected server.
        :return: a local representation of the requested document

        """
        doc = Document(self, id, data=default)

        try:
            await doc.fetch(discard_changes=True, rev=rev)
        except NotFoundError as e:
            if default is None:
                raise e

        return doc

    async def security(self) -> SecurityDocument:
        doc = SecurityDocument(self)
        await doc.fetch(discard_changes=True)
        return doc

    async def info(self) -> JsonDict:
        """Returns basic information about the database

        See also :ref:`GET /db<couchdb:api/db>`.

        :return: Description of returned object.
        :rtype: def

        """
        return await self._get()

    async def changes(
        self, last_event_id: Optional[str] = None, **params: Any
    ) -> AsyncGenerator[BaseChangeEvent, None]:
        """Listens for events made to documents of this database

        This will return :class:`~aiocouch.event.DeletedEvent` and
        :class:`~aiocouch.event.ChangedEvent` for deleted and modified
        documents, respectively.

        See also :ref:`/db/_changes<couchdb:api/db/changes>`.

        For convinience, the ``last-event-id`` parameter can also be passed
        as ``last_event_id``.

        """
        if last_event_id and "last-event-id" not in params:
            params["last-event-id"] = last_event_id

        async for json in self._changes(**params):
            if "deleted" in json and json["deleted"] is True:
                yield DeletedEvent(json=json)
            else:
                yield ChangedEvent(database=self, json=json)



import asyncio

from aiocouch import CouchDB


async def main_with() -> None:
    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:

        print((await couchdb.info())["version"])

        database = await couchdb["config"]

        async for doc in database.docs(["db-hta"]):
            print(doc)

    async with CouchDB(
        "http://localhost:5984", cookie="ZGVtb0B0b2x0ZWNrLmNvbT..."
    ) as couchdb:
        await couchdb["_users"]


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_with())



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import re
import sys

sys.path.insert(0, os.path.abspath(".."))

# -- Version setup -----------------------------------------------------------

_changelog_path = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "..", "CHANGELOG.md")
)
with open(_changelog_path, "r") as changelog:
    try:
        _version_info = re.match(
            r"# v(?P<major>\d+)\.(?P<minor>\d+)\.(?P<patch>\d+)",
            changelog.readline(),
        ).groupdict()
    except (OSError, IndexError):
        raise RuntimeError("Unable to determine version.")


# -- Project information -----------------------------------------------------

project = "aiocouch"
copyright = "2020, ZIH, Technische Universität Dresden"
author = "Mario Bielert"

# The short X.Y version
version = "{major}.{minor}".format(**_version_info)
# The full version
release = "{major}.{minor}.{patch}".format(**_version_info)

# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.coverage",
    "sphinx_autodoc_typehints",
    "sphinx.ext.intersphinx",
    "sphinxcontrib_trio",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

intersphinx_mapping = {
    "aiohttp": ("https://docs.aiohttp.org/en/stable/", None),
    "couchdb": ("https://docs.couchdb.org/en/stable/", None),
    "python": ("https://docs.python.org/3", None),
}


# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "alabaster"

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_theme_options = {
    "description": "An asynchronous CouchDB client library for asyncio and Python.",
    "canonical_url": "TODO",
    "github_user": "metricq",
    "github_repo": "aiocouch",
    "github_button": True,
    "github_type": "star",
    "github_banner": True,
    "fixed_sidebar": True,
    "codecov_button": True,
}



from typing import AsyncGenerator, NoReturn, cast

import pytest
from aiohttp.client import RequestInfo
from aiohttp.client_exceptions import ClientResponseError

from aiocouch.couchdb import JsonDict
from aiocouch.exception import (
    BadRequestError,
    ConflictError,
    ExpectationFailedError,
    ForbiddenError,
    NotFoundError,
    PreconditionFailedError,
    UnauthorizedError,
    UnsupportedMediaTypeError,
    generator_raises,
    raises,
)

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


class CustomError(Exception):
    pass


class DummyEndpoint:
    @property
    def endpoint(self) -> str:
        return "endpoint"

    @raises(400, "bad thing")
    async def raise_bad_request(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=400)

    @raises(401, "bad thing")
    async def raise_unauthorized(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=401)

    @raises(403, "bad thing")
    async def raise_forbidden(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=403)

    @raises(404, "bad thing")
    async def raise_not_found(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=404)

    @raises(409, "bad thing")
    async def raise_conflict(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=409)

    @raises(412, "bad thing")
    async def raise_precondition_failed(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=412)

    @raises(415, "bad thing")
    async def raise_unsupported_media(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=415)

    @raises(417, "bad thing")
    async def raise_expectation_failed(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=417)

    @raises(500, "bad thing", CustomError)
    async def raise_custom(self) -> NoReturn:
        raise ClientResponseError(cast(RequestInfo, None), (), status=500)

    @generator_raises(400, "bad thing")
    async def raise_in_generator(self) -> AsyncGenerator[JsonDict, None]:
        raise ClientResponseError(cast(RequestInfo, None), (), status=400)
        yield None


async def test_raises() -> None:
    dummy = DummyEndpoint()
    with pytest.raises(BadRequestError):
        await dummy.raise_bad_request()

    with pytest.raises(UnauthorizedError):
        await dummy.raise_unauthorized()

    with pytest.raises(ForbiddenError):
        await dummy.raise_forbidden()

    with pytest.raises(NotFoundError):
        await dummy.raise_not_found()

    with pytest.raises(ConflictError):
        await dummy.raise_conflict()

    with pytest.raises(PreconditionFailedError):
        await dummy.raise_precondition_failed()

    with pytest.raises(UnsupportedMediaTypeError):
        await dummy.raise_unsupported_media()

    with pytest.raises(ExpectationFailedError):
        await dummy.raise_expectation_failed()

    with pytest.raises(CustomError):
        await dummy.raise_custom()

    with pytest.raises(BadRequestError):
        async for _ in dummy.raise_in_generator():
            pass



import pytest

from aiocouch import CouchDB, Database

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_list_databases(couchdb: CouchDB) -> None:
    dbs = await couchdb.keys()
    assert "aiocouch_test_fixture_database" not in dbs


async def test_list_database(couchdb: CouchDB, database: Database) -> None:
    dbs = await couchdb.keys()
    assert "aiocouch_test_fixture_database" in dbs


async def test_create_delete_database(couchdb: CouchDB) -> None:
    database = await couchdb.create("aiocouch_test_fixture_database2")

    dbs = await couchdb.keys()
    assert "aiocouch_test_fixture_database2" in dbs

    await database.delete()

    dbs = await couchdb.keys()
    assert "aiocouch_test_fixture_database2" not in dbs


async def test_create_for_existing(couchdb: CouchDB, database: Database) -> None:
    from aiocouch import PreconditionFailedError

    with pytest.raises(PreconditionFailedError):
        await couchdb.create(database.id)


async def test_create_for_existing_ok(couchdb: CouchDB, database: Database) -> None:
    await couchdb.create(database.id, exists_ok=True)


async def test_get_for_existing(couchdb: CouchDB, database: Database) -> None:
    await couchdb[database.id]


async def test_get_for_non_existing(couchdb: CouchDB, database: Database) -> None:
    from aiocouch import NotFoundError

    with pytest.raises(NotFoundError):
        await couchdb[database.id + "not_existing"]


async def test_get_info(couchdb: CouchDB) -> None:
    await couchdb.info()


async def test_get_database_info(database: Database) -> None:
    await database.info()



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from types import TracebackType
from typing import Any, AsyncGenerator, Dict, List, Optional, Type, cast

from deprecated.sphinx import deprecated

from . import database
from .document import Document

JsonDict = Dict[str, Any]


class BulkOperation:
    """A context manager for bulk operation. This operation allows to
    write many documents in one request.

    Bulk operations use the :ref:`_bulk_docs<couchdb:api/db/bulk_docs>`
    endpoint of the database.

    To populate the list of written documents, use the :meth:`append`
    method.

    :param database: The database used in the bulk operation
    """

    def __init__(self, database: "database.Database"):
        self._database = database

        self.response: Optional[List[JsonDict]] = None
        """The resulting JSON response of the :ref:`_bulk_docs<couchdb:api/db/bulk_docs>`
        request. Refer to the CouchDB documentation for the contents.

        Only available after the context manager has finished without a
        passing exception."""

        self.ok: Optional[List[Document]] = None
        """The list of all :class:`~aiocouch.document.Document` instances that there
        successfully saved to server.

        Only available after the context manager has finished without a
        passing exception."""

        self.error: Optional[List[Document]] = None
        """The list of all :class:`~aiocouch.document.Document` instances that could not
        be saved to server.

        Only available after the context manager has finished without a
        passing exception."""

        self._docs: List[Document] = []

    async def __aenter__(self) -> "BulkOperation":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is not None:
            return

        # @VTTI: Yes, we actually need doc._data and not doc.data here
        docs = [doc._data for doc in self._docs if doc._dirty_cache]

        if docs:
            self.response = cast(List[JsonDict], await self._database._bulk_docs(docs))
        else:
            self.response = []

        self.ok = []
        self.error = []

        result_docs = [doc for doc in self._docs if doc._dirty_cache]
        for status, doc in zip(self.response, result_docs):
            assert status["id"] == doc.id
            if "ok" in status:
                doc._update_rev_after_save(status)
                self.ok.append(doc)
            else:
                self.error.append(doc)

    async def __aiter__(self) -> AsyncGenerator[Document, None]:
        """An iterator that yields Document instances that are part of this bulk operation.

        :return: Every :class:`~aiocouch.document.Document` instance that will be affected by this operation
        :rtype: AsyncGenerator[Document, None]
        """
        for doc in self._docs:
            yield doc

    def create(self, id: str, data: Optional[JsonDict] = None) -> Document:
        """Create a new document as part of the bulk operation

        :param id: the id of the document
        :param data: the inital data used to set the body of the document, defaults to None
        :raises ValueError: if the provided document id is already part of the bulk operation
        :return: a Document instance reference the newly created document
        """
        if any(id == d.id for d in self._docs):
            raise ValueError(
                f"There is already another Document instance for {id} part of the BulkOperation"
            )

        doc = Document(self._database, id, data=data)
        self._docs.append(doc)

        return doc

    # Mypy isn't smart enough, see: https://github.com/python/mypy/issues/1362
    @property  # type:ignore
    @deprecated(version="2.1.0", reason="Use the response property instead.")
    def status(self) -> Optional[List[JsonDict]]:  # pragma: no cover
        return self.response

    @deprecated(
        version="2.1.0", reason="Use append(doc) instead. It just makes more sense."
    )
    def update(self, doc: Document) -> Document:  # pragma: no cover
        """Add a document to this bulk operation.

        :param doc: the document that should be stored as part of the bulk operation
        :raises ValueError: if the provided document instance is already part of the bulk operation
        :return: the provided document
        """
        return self.append(doc)

    def append(self, doc: Document) -> Document:
        """Add a document to this bulk operation.

        :param doc: the document that should be stored as part of the bulk operation
        :raises ValueError: if the provided document instance is already part of the bulk operation
        :return: the provided document
        """
        if any(doc.id == d.id for d in self._docs):
            raise ValueError(
                f"There is already another Document instance for {doc.id} part of the BulkOperation"
            )

        self._docs.append(doc)
        return doc


class BulkCreateOperation(BulkOperation):
    """A context manager for bulk creation operations. This operation allows to
    write many documents in one request.

    Bulk operations use the :ref:`_bulk_docs<couchdb:api/db/bulk_docs>`
    endpoint of the database.

    :param database: The database used in the bulk operation
    :param ids: a list of ids of the involved documents, defaults to []
    """

    def __init__(self, database: "database.Database", ids: List[str] = []):
        super().__init__(database=database)
        self._ids = ids

    async def __aenter__(self) -> "BulkCreateOperation":
        self._docs.extend([Document(self._database, id) for id in self._ids])
        return self


class BulkUpdateOperation(BulkOperation):
    """A context manager for bulk update of documents. In particular, for every provided
    id, a :class:`~aiocouch.document.Document` instance is provided. The data is fetched
    using the :class:`~aiocouch.view.AllDocsView` with a minimal amount of requests.

    :param database: The database of the bulk operation
    :param ids: list of document ids
    :param create: If ``True``, every document contained in `ids` that doesn't
            exist, will be represented by an empty
            :class:`~aiocouch.document.Document` instance.
    """

    def __init__(
        self, database: "database.Database", ids: List[str] = [], create: bool = False
    ):
        super().__init__(database=database)
        self._ids = ids
        self._create = create

    async def __aenter__(self) -> "BulkUpdateOperation":
        self._docs.extend(
            [doc async for doc in self._database.docs(self._ids, create=self._create)]
        )
        return self



import asyncio

from aiocouch import CouchDB


async def main_with() -> None:
    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:
        db = await couchdb["recipes"]
        doc = await db["apple_pie"]
        print(doc["incredients"])

        new_doc = await db.create(
            "cookies", data={"title": "Grandma's cookies", "rating": "★★★★★"}
        )
        await new_doc.save()


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_with())



import asyncio
import uuid

from aiocouch import CouchDB
from aiocouch.event import ChangedEvent, DeletedEvent


async def main_with() -> None:

    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:
        database = await couchdb["test"]

        async for event in database.changes(include_docs=True):
            if isinstance(event, DeletedEvent):
                print(f"The document {event.id} was deleted.")
            elif isinstance(event, ChangedEvent):
                print(f"The document {event.id} was saved as {event.rev}")


if __name__ == "__main__":
    loop = asyncio.get_event_loop()
    loop.run_until_complete(main_with())



# Copyright (c) 2021, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, Dict, List, Union

JsonDict = Dict[str, Any]
JsonList = List[Any]
Json = Union[JsonDict, JsonList]



import pytest

from aiocouch import CouchDB
from aiocouch.database import Database

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_security_missing_privileges(
    database: Database, couchdb_with_user_access: CouchDB
) -> None:
    from aiocouch import ForbiddenError

    sec2 = await database.security()
    sec2.add_member("aiocouch_test_user")
    await sec2.save()

    db = await couchdb_with_user_access["aiocouch_test_fixture_database"]
    sec = await db.security()

    sec.add_member("foobert")

    with pytest.raises(ForbiddenError):
        await sec.save()

    sec2.add_admin("aiocouch_test_user")
    await sec2.save()

    await sec.save()


async def test_get_missing_privileges(
    filled_database: Database, couchdb_with_user_access: CouchDB
) -> None:
    from aiocouch import ForbiddenError

    sec = await filled_database.security()
    sec.add_member("foobert")
    await sec.save()

    with pytest.raises(ForbiddenError):
        await couchdb_with_user_access["aiocouch_test_fixture_database"]


async def test_get_valid_privileges(
    filled_database: Database, couchdb_with_user_access: CouchDB
) -> None:
    from aiocouch import ForbiddenError

    sec = await filled_database.security()
    sec.add_member("aiocouch_test_user")
    await sec.save()

    database = await couchdb_with_user_access["aiocouch_test_fixture_database"]
    doc = await database["foo"]
    doc["bear"] = "brown"

    sec.remove_member("aiocouch_test_user")
    sec.add_member("foobert")
    await sec.save()

    with pytest.raises(ForbiddenError):
        await doc.save()

    with pytest.raises(ForbiddenError):
        await couchdb_with_user_access["aiocouch_test_fixture_database"]

    sec.add_member_role("aiocouch_test_role")
    await sec.save()

    database = await couchdb_with_user_access["aiocouch_test_fixture_database"]
    await database["foo"]

    await doc.save()



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from types import TracebackType
from typing import Any, Dict, List, Optional, Type

from .database import Database
from .exception import NotFoundError, PreconditionFailedError
from .remote import RemoteServer

JsonDict = Dict[str, Any]


class CouchDB:
    """CouchDB Server Connection Session

    The

    :param str server: URL of the CouchDB server
    :param str user: user used for authentication
    :param str password: password for authentication
    :param str cookie: The session cookie used for authentication
    :param Any kwargs: Any other kwargs are passed to :class:`aiohttp.ClientSession`

    """

    def __init__(self, *args: Any, **kwargs: Any):
        self._server = RemoteServer(*args, **kwargs)

    async def __aenter__(self) -> "CouchDB":
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def check_credentials(self) -> None:
        """Check the provided credentials.

        :raises ~aiocouch.UnauthorizedError: if provided credentials aren't valid

        """
        await self._server._check_session()

    async def close(self) -> None:
        """Closes the connection to the CouchDB server"""
        await self._server.close()

    async def create(
        self, id: str, exists_ok: bool = False, **kwargs: Any
    ) -> "Database":
        """Creates a new database on the server

        :raises ~aiocouch.PreconditionFailedError: if the database already
            exists and ``exists_ok`` is ``False``

        :param id: the identifier of the database
        :param exists_ok: If ``True``, don't raise if the database exists
        :return: Returns a representation for the created database

        """
        db = Database(self, id)
        if not await db._exists():
            await db._put(**kwargs)
            return db
        elif exists_ok:
            return db
        else:
            raise PreconditionFailedError(f"The database '{id}' does already exist.")

    async def __getitem__(self, id: str) -> "Database":
        """Returns a representation for the given database identifier

        :raises ~aiocouch.NotFoundError: if the database does not exist

        :param id: The identifier of the database
        :return: The representation of the database

        """
        db = Database(self, id)

        if not await db._exists():
            raise NotFoundError(f"The database '{id}' does not exist.")

        return db

    async def keys(self, **params: Any) -> List[str]:
        """Returns all database names

        :return: A list containing the names of all databases on the server

        """
        return await self._server._all_dbs(**params)

    async def info(self) -> JsonDict:
        """Returns the meta information about the connected CouchDB server.

        See also :ref:`GET /<couchdb:api/server/root>`

        :return: A dict containing the response json.

        """
        return await self._server._info()



import asyncio
import timeit

from aiocouch import CouchDB


async def main_with() -> None:
    async with CouchDB(
        "http://localhost:5984", user="admin", password="admin"
    ) as couchdb:

        database = await couchdb["unfun"]

        # selector = {
        #     "_id": {
        #      "$regex": "^taurus\\.taurusi5172.*$"
        #     }
        # }
        #
        # async for doc in database.find(selector=selector):
        #     print(doc)

        # async for doc in database.docs():
        #     if doc.id.startswith("taurus.taurusi5172."):
        #         print(doc)

        # async for doc in database.docs(prefix="taurus.taurusi5172."):
        #     print(doc)

        async for key in database.all_docs.akeys(prefix="x"):
            print(key)

        print("------")

        async for key in database.view("view", "new-view").akeys(prefix="x"):
            print(key)


if __name__ == "__main__":
    loop = asyncio.get_event_loop()

    repeat = 10

    print(
        timeit.timeit(lambda: loop.run_until_complete(main_with()), number=repeat)
        / repeat
    )

    # loop.run_until_complete(main_with())



# Copyright (c) 2021, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.


from typing import cast

from attr import dataclass

from . import database as db
from . import document
from .typing import JsonDict


@dataclass
class BaseChangeEvent:
    """The base event for shared properties"""

    json: JsonDict
    """The raw data of the event as JSON"""

    @property
    def id(self) -> str:
        """Returns the id of the document"""
        return cast(str, self.json["id"])

    @property
    def rev(self) -> str:
        """Returns the new rev of the document"""
        return cast(str, self.json["changes"][0]["rev"])

    @property
    def sequence(self) -> str:
        """Returns the sequence identifier of the event"""
        return cast(str, self.json["seq"])


class DeletedEvent(BaseChangeEvent):
    """This event denotes that the document got deleted"""

    pass


@dataclass
class ChangedEvent(BaseChangeEvent):
    """This event denotes that the document got modified"""

    database: "db.Database"
    """The database for reference"""

    async def doc(self) -> "document.Document":
        """Returns the document after the change

        If the ``include_docs`` was set, this will use the data provided in the received event.
        Otherwise, the document is fetched from the server.
        """
        try:
            # if in the request include_docs was given, we can create the
            # document on the spot...
            return document.Document(
                self.database, self.json["doc"]["_id"], self.json["doc"]
            )
        except KeyError:
            # ...otherwise, we fetch the document contents from the server
            return await self.database.get(self.id, rev=self.rev)



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import json
from contextlib import suppress
from types import TracebackType
from typing import Any, ItemsView, KeysView, List, Optional, Type, ValuesView, cast

from deprecated import deprecated

from . import database
from .attachment import Attachment
from .exception import ConflictError, ForbiddenError, NotFoundError, raises
from .remote import RemoteDocument
from .typing import JsonDict


class Document(RemoteDocument):
    """A local representation for the referenced CouchDB document

    An instance of this class represents a local copy of the document data on the
    server. This class behaves like a dict containing the document data and allows to
    :func:`~aiocouch.document.Document.fetch` and
    :func:`~aiocouch.document.Document.save` documents. For details about the dict-like
    interface, please refer to the :ref:`Python manual <python:typesmapping>`.

    Constructing an instance of this class does not cause any network requests.

    :ivar id: the id of the document

    :param `~aiocouch.database.Database` database: The database of the document
    :param id: the id of the document
    :param data: the inital data used to set the body of the document

    """

    def __init__(
        self, database: "database.Database", id: str, data: Optional[JsonDict] = None
    ):
        super().__init__(database, id)
        self._data: JsonDict = data if data is not None else {}
        if not isinstance(self._data, dict):
            raise TypeError("data parameter must be a dict object")
        self._data["_id"] = id
        self._data_hash: Optional[int] = None

    async def __aenter__(self) -> "Document":
        with suppress(NotFoundError):
            await self.fetch(discard_changes=True)
        return self

    async def __aexit__(
        self,
        exc_type: Optional[Type[BaseException]],
        exc_value: Optional[BaseException],
        traceback: Optional[TracebackType],
    ) -> None:
        if exc_type is None:
            await self.save()

    def _update_hash(self) -> None:
        self._data_hash = hash(json.dumps(self._data, sort_keys=True))

    @property
    def _fresh(self) -> bool:
        return len(self._data) == 1 and "_id" in self._data

    @property
    def _dirty_cache(self) -> bool:
        return self._data_hash is None or self._data_hash != hash(
            json.dumps(self._data, sort_keys=True)
        )

    async def fetch(
        self, discard_changes: bool = False, *, rev: Optional[str] = None
    ) -> None:
        """Retrieves the document data from the server

        Fetching the document will retrieve the data from the server using a network
        request and update the local data.

        :raises ~aiocouch.ConflictError: if the local data has changed without saving
        :raises ~aiocouch.BadRequestError: if the given rev is invalid or missing

        :param discard_changes: If set to `True`, the local data object will the
            overridden with the retrieved content. If the local data was changed, no
            exception will be raised.
        :param rev: The requested rev of the document. The requested rev might not
            or not anymore exist on the connected server.

        """
        if self._dirty_cache and not (discard_changes or self._fresh):
            raise ConflictError(
                f"Cannot fetch document '{self.id}' from server, "
                "as the local cache has unsaved changes."
            )

        if rev:
            self._update_cache(await self._get(rev=rev))
        else:
            self._update_cache(await self._get())

    async def save(self) -> None:
        """Saves the current state to the CouchDB server

        :raises ~aiocouch.ConflictError: if the local revision is different from the
            server. See `Conflict handling`_.

        """
        if self._dirty_cache:
            data = await self._put(self._data)
            self._update_rev_after_save(data)

    async def delete(self, discard_changes: bool = False) -> None:
        """Deletes the document from the server

        Calling this method deletes the local data and the document on the server.
        Afterwards, the instance can be filled with new data and call :meth:`.save`
        again.

        :raises ~aiocouch.ConflictError: if the local data has changed without saving
        :raises ~aiocouch.ConflictError: if the local revision is different from the
            server. See `Conflict handling`_.

        """
        if self._dirty_cache and not discard_changes:
            raise ConflictError(
                f"Cannot delete document '{self.id}' from server, as the local cache "
                "has unsaved changes."
            )
        self._update_cache(await self._delete(rev=self["_rev"]))

    async def copy(self, new_id: str) -> "Document":
        """Create a copy of the document on the server

        Creates a new document with the data currently stored on the server.

        :param new_id: the id of the new document
        :return: an instance for the new document after it was copied

        """
        await self._copy(new_id)

        return await self._database[new_id]

    @property
    def rev(self) -> Optional[str]:
        """Allows to set and get the local revision

        If the local document wasn't fetched or saved, this is ``None``.

        """
        try:
            return cast(str, self._data["_rev"])
        except KeyError:
            return None

    @rev.setter
    def rev(self, new_rev: str) -> None:
        if not isinstance(new_rev, str):
            raise TypeError("Revision must be a string.")
        self._data["_rev"] = new_rev

    @property
    def data(self) -> Optional[JsonDict]:
        """Returns the document as a dict

        If :func:`~aiocouch.document.Document.exists` is ``False``, this function returns ``None``.

        This method does not perform a network request.

        :return: Returns the data of the document or ``None``

        """
        return self._data if self.exists else None

    @property
    def exists(self) -> bool:
        """Denotes whether the document exists

        A document exists, if an existing was :func:`~aiocouch.document.Document.fetch` ed from
        the server and retrieved data doesn't contain the `_deleted` field. Or a new document
        was saved using :func:`~aiocouch.document.Document.save`.

        This method does not perform a network request.

        :return: ``True`` if the document exists, ``False`` overwise

        """
        return "_rev" in self and "_deleted" not in self

    @deprecated(
        version="1.1.0", reason="This method is a misnomer. Use info() instead."
    )
    async def fetch_info(self) -> JsonDict:  # pragma: no cover
        return await self._info()

    async def info(self) -> JsonDict:
        """Returns a short information about the document.

        This method sends a request to the server to retrieve the current status.

        :raises ~aiocouch.NotFoundError: if the document does not exist on the server

        :return: A dict containing the id and revision of the document on the server

        """
        return await self._info()

    def _update_rev_after_save(self, data: JsonDict) -> None:
        with suppress(KeyError):
            self._data["_rev"] = data["rev"]
        self._update_hash()

    def _update_cache(self, new_cache: JsonDict) -> None:
        self._data = new_cache
        self._update_hash()

    def __getitem__(self, key: str) -> Any:
        return self._data[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._data[key] = value

    def __delitem__(self, key: str) -> None:
        del self._data[key]

    def __contains__(self, key: str) -> bool:
        return key in self._data

    def update(self, data: JsonDict) -> None:
        self._data.update(data)

    def items(self) -> ItemsView[str, Any]:
        return self._data.items()

    def keys(self) -> KeysView[str]:
        return self._data.keys()

    def values(self) -> ValuesView[Any]:
        return self._data.values()

    def get(self, key: str, default: Optional[Any] = None) -> Any:
        return self._data.get(key, default)

    def setdefault(self, key: str, default: Optional[Any] = None) -> Any:
        return self._data.setdefault(key, default)

    def attachment(self, id: str) -> "Attachment":
        """Returns the attachment object

        The attachment object is returned, but this method doesn't actually fetch any
        data from the server. Use
        :meth:`~aiocouch.attachment.Attachment.fetch()` and
        :meth:`~aiocouch.attachment.Attachment.save()`, respectively.

        :param id: the id of the attachment
        :return: Returns the attachment object

        """
        return Attachment(self, id)

    def __repr__(self) -> str:
        return json.dumps(self._data, indent=2)


class SecurityDocument(Document):
    def __init__(self, database: "database.Database"):
        super().__init__(database, "_security")
        del self._data["_id"]

    async def __aenter__(self) -> "SecurityDocument":
        return cast(SecurityDocument, await super().__aenter__())

    @property
    def members(self) -> Optional[List[str]]:
        try:
            return cast(List[str], self["members"]["names"])
        except KeyError:
            return None

    @property
    def member_roles(self) -> Optional[List[str]]:
        try:
            return cast(List[str], self["members"]["roles"])
        except KeyError:
            return None

    @property
    def admins(self) -> Optional[List[str]]:
        try:
            return cast(List[str], self["admins"]["names"])
        except KeyError:
            return None

    @property
    def admin_roles(self) -> Optional[List[str]]:
        try:
            return cast(List[str], self["admins"]["roles"])
        except KeyError:
            return None

    def add_member(self, member: str) -> None:
        members = self.setdefault("members", {})
        names = members.setdefault("names", [])
        if member not in names:
            names.append(member)

    def add_member_role(self, role: str) -> None:
        members = self.setdefault("members", {})
        roles = members.setdefault("roles", [])
        if role not in roles:
            roles.append(role)

    def remove_member(self, member: str) -> None:
        try:
            self["members"]["names"].remove(member)
        except (ValueError, KeyError) as e:
            raise KeyError(
                f"The user '{member}' isn't a member of the database '{self._database.id}'"
            ) from e

    def remove_member_role(self, role: str) -> None:
        try:
            self["members"]["roles"].remove(role)
        except (ValueError, KeyError) as e:
            raise KeyError(
                f"The role '{role}' isn't a member role of the database '{self._database.id}'"
            ) from e

    def add_admin(self, admin: str) -> None:
        admins = self.setdefault("admins", {})
        names = admins.setdefault("names", [])
        if admin not in names:
            names.append(admin)

    def add_admin_role(self, role: str) -> None:
        admins = self.setdefault("admins", {})
        roles = admins.setdefault("roles", [])
        if role not in roles:
            roles.append(role)

    def remove_admin(self, admin: str) -> None:
        try:
            self["admins"]["names"].remove(admin)
        except (ValueError, KeyError) as e:
            raise KeyError(
                f"The user '{admin}' isn't an admin of the database '{self._database.id}'"
            ) from e

    def remove_admin_role(self, role: str) -> None:
        try:
            self["admins"]["roles"].remove(role)
        except (ValueError, KeyError) as e:
            raise KeyError(
                f"The role '{role}' isn't an admin role of the database '{self._database.id}'"
            ) from e

    @raises(500, "You are not a database or server admin", ForbiddenError)
    async def save(self) -> None:
        await super().save()



# Copyright (c) 2019, ZIH,
# Technische Universitaet Dresden,
# Federal Republic of Germany
#
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without modification,
# are permitted provided that the following conditions are met:
#
#     * Redistributions of source code must retain the above copyright notice,
#       this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above copyright notice,
#       this list of conditions and the following disclaimer in the documentation
#       and/or other materials provided with the distribution.
#     * Neither the name of metricq nor the names of its contributors
#       may be used to endorse or promote products derived from this software
#       without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
# EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
# PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
# PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
# LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
# NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
# SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from typing import Any, AsyncGenerator, Dict, List, Optional, Tuple, cast

from . import database
from .document import Document
from .exception import NotFoundError
from .remote import RemoteView

JsonDict = Dict[str, Any]


class View(RemoteView):
    def __init__(
        self, database: "database.Database", design_doc: Optional[str], id: str
    ):
        super().__init__(database, design_doc, id)

    @property
    def prefix_sentinal(self) -> str:
        return "\uffff"

    async def get(self, **params: Any) -> AsyncGenerator[JsonDict, None]:
        result_chunk = await self._get(**params)

        for res in result_chunk["rows"]:
            yield res

    async def post(
        self, ids: List[str], create: bool = False, **params: Any
    ) -> AsyncGenerator[JsonDict, None]:
        result_chunk = await self._post(ids, **params)

        for res in result_chunk["rows"]:
            yield res

    async def ids(
        self,
        keys: Optional[List[str]] = None,
        prefix: Optional[str] = None,
        **params: Any,
    ) -> AsyncGenerator[str, None]:
        if prefix is not None:
            params["startkey"] = f'"{prefix}"'
            params["endkey"] = f'"{prefix}{self.prefix_sentinal}"'

        if keys is None:
            async for res in self.get(**params):
                if "error" not in res:
                    yield res["id"]
        else:
            async for res in self.post(keys, **params):
                if "error" not in res:
                    yield res["id"]

    async def akeys(self, **params: Any) -> AsyncGenerator[str, None]:
        async for res in self.get(**params):
            yield res["key"]

    async def aitems(self, **params: Any) -> AsyncGenerator[Tuple[str, Any], None]:
        async for res in self.get(**params):
            yield cast(str, res["key"]), res["value"]

    async def avalues(self, **params: Any) -> AsyncGenerator[Any, None]:
        async for res in self.get(**params):
            yield res["value"]

    async def docs(
        self,
        ids: Optional[List[str]] = None,
        create: bool = False,
        prefix: Optional[str] = None,
        include_ddocs: bool = False,
        **params: Any,
    ) -> AsyncGenerator[Document, None]:
        params["include_docs"] = True
        if prefix is None:
            if ids is None:
                iter = self.get(**params)
            else:
                iter = self.post(ids, **params)
        else:
            if ids is not None or create:
                raise ValueError(
                    "prefix cannot be used together with ids or create parameter"
                )

            params["startkey"] = f'"{prefix}"'
            params["endkey"] = f'"{prefix}{self.prefix_sentinal}"'

            iter = self.get(**params)

        async for res in iter:
            if "error" not in res and res["doc"] is not None:
                if res["id"].startswith("_design/") and not include_ddocs:
                    continue
                doc = Document(self._database, res["id"])
                doc._update_cache(res["doc"])
                yield doc
            elif create:
                doc = Document(self._database, res["key"])
                yield doc
            else:
                raise NotFoundError(
                    f"The document '{res['key']}' does not exist in the database "
                    f"{self._database.id}."
                )


class AllDocsView(View):
    def __init__(self, database: "database.Database"):
        super().__init__(database, None, "_all_docs")

    @property
    def endpoint(self) -> str:
        return f"/{self._database.id}/_all_docs"

    @property
    def prefix_sentinal(self) -> str:
        return chr(0x10FFFE)



import pytest

from aiocouch.database import Database
from aiocouch.document import Document

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def test_update_docs_on_empty(database: Database) -> None:
    async with database.update_docs([]) as bulk:
        pass

    assert bulk._docs == []
    keys = [key async for key in database.akeys()]
    assert len(keys) == 0


async def test_update_docs_creating(database: Database) -> None:
    async with database.update_docs(["foobar"], create=True):
        pass

    keys = [key async for key in database.akeys()]

    assert len(keys) == 1
    assert sorted(keys) == ["foobar"]


async def test_update_docs_creating_not_ok(database: Database) -> None:
    with pytest.raises(KeyError):
        async with database.update_docs(["foobar"]):
            pass


async def test_update_docs(database: Database) -> None:
    async with database.update_docs(["foo", "baz"], create=True) as bulk:
        async for doc in bulk:
            doc["llama"] = "awesome"

    keys = [key async for key in database.akeys()]

    assert len(keys) == 2
    assert sorted(keys) == ["baz", "foo"]

    async for doc in database.values():
        assert "llama" in doc
        assert doc["llama"] == "awesome"


async def test_update_docs_no_change(filled_database: Database) -> None:
    async with filled_database.update_docs(["foo", "baz"]) as bulk:
        pass

    assert bulk.response == []


async def test_update_dont_crash_on_pristine_doc(filled_database: Database) -> None:
    doc = await filled_database["foo"]
    doc["llama"] = "awesome"
    await doc.save()

    async with filled_database.update_docs(["foo", "baz"], create=True) as bulk:
        async for doc in bulk:
            doc["llama"] = "awesome"


async def test_update_docs_for_deleted(filled_database: Database) -> None:
    doc = await filled_database["foo"]
    await doc.delete()

    async with filled_database.update_docs(["foo"], create=True) as bulk:
        async for doc in bulk:
            doc["llama"] = "awesome"

    doc = await filled_database["foo"]

    assert "_deleted" not in doc
    assert "_rev" in doc
    assert doc["_rev"].startswith("3-")
    assert doc["llama"] == "awesome"


async def test_update_docs_for_errored(filled_database: Database) -> None:
    doc = await filled_database["foo"]
    doc["something"] = 42
    async with filled_database.update_docs(["foo", "baz"]) as bulk:
        # provoke a conflict for document foo
        await doc.save()

        async for doc in bulk:
            doc["thing"] = 42

    assert bulk.response is not None
    assert len(bulk.response) == 2

    assert bulk.ok is not None
    assert len(bulk.ok) == 1
    doc_ok = bulk.ok[0]
    assert doc_ok.id == "baz"
    assert doc_ok["thing"] == 42
    assert doc_ok["_rev"].startswith("2-")

    assert bulk.error is not None
    assert len(bulk.error) == 1
    doc_err = bulk.error[0]
    assert doc_err.id == "foo"
    assert doc_err["_rev"].startswith("1-")
    assert "something" not in doc_err


async def test_create_docs_with_ids(database: Database) -> None:
    async with database.create_docs(["foo", "baz"]) as bulk:
        pass

    assert bulk.response is not None
    assert len(bulk.response) == 2

    keys = [key async for key in database.akeys()]

    assert len(keys) == 2
    assert sorted(keys) == ["baz", "foo"]


async def test_create_docs_with_create(database: Database) -> None:
    async with database.create_docs() as bulk:
        bulk.create("foo", data={"counter": 42})
        bulk.create("baz")

        with pytest.raises(ValueError):
            bulk.create("foo")

    assert bulk.response is not None
    assert len(bulk.response) == 2

    keys = [key async for key in database.akeys()]

    assert len(keys) == 2
    assert sorted(keys) == ["baz", "foo"]

    foo = await database["foo"]
    assert "counter" in foo
    assert foo["counter"] == 42


async def test_create_docs_with_create_duplicate(database: Database) -> None:
    async with database.create_docs() as bulk:
        foo = bulk.create("foo")

        # DO NOT DO THIS! This is just using the private interface to test conflict handling.
        assert bulk._docs is not None
        bulk._docs.append(foo)

    assert bulk.response is not None
    assert len(bulk.response) == 2

    assert "ok" in bulk.response[0]
    assert "error" in bulk.response[1]
    assert bulk.response[1]["error"] == "conflict"

    keys = [key async for key in database.akeys()]

    assert len(keys) == 1
    assert sorted(keys) == ["foo"]


async def test_create_docs_mixed(database: Database) -> None:
    async with database.create_docs(["foo"]) as bulk:
        bulk.create("baz")

    assert bulk.response is not None
    assert len(bulk.response) == 2

    keys = [key async for key in database.akeys()]

    assert len(keys) == 2
    assert sorted(keys) == ["baz", "foo"]


async def test_create_docs_for_existing(filled_database: Database) -> None:
    async with filled_database.create_docs(["new", "foo"]) as bulk:
        bulk.create("baz")

    assert bulk.response is not None
    assert bulk.error is not None
    assert bulk.ok is not None

    assert len(bulk.response) == 3
    assert len(bulk.ok) == 1
    assert len(bulk.error) == 2
    assert bulk.response[1]["error"] == "conflict"
    assert bulk.response[2]["error"] == "conflict"


async def test_update_external_documents(filled_database: Database) -> None:
    foo = await filled_database.get("foo")

    assert "zebras" not in foo

    async with filled_database.update_docs() as bulk:
        foo["zebras"] = "awesome 🦓"
        bulk.append(foo)

        with pytest.raises(ValueError):
            bulk.append(foo)

    foo = await filled_database.get("foo")
    assert foo["zebras"] == "awesome 🦓"


async def test_no_bulk_request_on_exception(database: Database, doc: Document) -> None:
    from aiocouch.bulk import BulkOperation

    doc["zebras"] = "awesome 🦓"

    with pytest.raises(Exception):
        async with BulkOperation(database=database) as bulk:
            bulk.append(doc)

            # simulate an error
            raise Exception()

    assert bulk.error is None
    assert bulk.ok is None
    assert bulk.response is None

    # check that the changes were not send to the server
    doc2 = await database.create(doc.id)
    assert "zebras" not in doc2



def test_stringify_params() -> None:
    from aiocouch.remote import _stringify_params

    assert _stringify_params(None) is None
    assert _stringify_params({}) == {}

    params = {"foo": "bar", "baz": True, "boo": False, "foz": None}

    assert _stringify_params(params) == {"foo": "bar", "baz": "true", "boo": "false"}



import asyncio
from typing import Any, List, cast

import pytest

from aiocouch.database import Database
from aiocouch.document import Document
from aiocouch.event import BaseChangeEvent, ChangedEvent, DeletedEvent

# All test coroutines will be treated as marked.
pytestmark = pytest.mark.asyncio


async def listen_for_first_change(database: Database, **kwargs: Any) -> BaseChangeEvent:
    async for event in database.changes(**kwargs, feed="continuous", since="now"):
        return event

    assert False


async def listen_for_changes(
    database: Database, **kwargs: Any
) -> List[BaseChangeEvent]:
    return [event async for event in database.changes(**kwargs)]


async def test_changed_event_for_new_document(database: Database) -> None:
    async def create_doc(database: Database) -> Document:
        await asyncio.sleep(0.1)
        doc = await database.create("foo", data={"Zebras": "are pants"})
        await doc.save()

        return doc

    results = await asyncio.gather(
        listen_for_first_change(database), create_doc(database)
    )

    event = results[0]
    doc = results[1]

    assert isinstance(event, ChangedEvent)
    assert event.id == "foo"
    assert event.rev == doc.rev

    doc2 = await event.doc()
    assert doc2.id == "foo"
    assert doc2.rev == event.rev
    assert "Zebras" in doc2


async def test_changed_event_for_existing_doc(filled_database: Database) -> None:
    async def update_doc(doc: Document) -> None:
        await asyncio.sleep(0.1)
        doc["Zebras"] = "are black with white stripes"
        await doc.save()

    doc = await filled_database["foo"]

    results = await asyncio.gather(
        listen_for_first_change(filled_database), update_doc(doc)
    )

    event = results[0]

    assert isinstance(event, ChangedEvent)
    assert event.id == "foo"
    assert event.rev == doc.rev

    doc2 = await event.doc()
    assert doc2.id == "foo"
    assert doc2.rev == event.rev
    assert "Zebras" in doc2


async def test_polling(filled_database: Database) -> None:
    old_events = await listen_for_changes(filled_database, feed="normal")
    last_event_id = old_events[-1].sequence if len(old_events) else None

    doc = await filled_database["foo"]
    doc["Zebras"] = "are black with white stripes"
    await doc.save()

    events = await listen_for_changes(
        filled_database, feed="normal", last_event_id=last_event_id
    )

    assert len(events) == 1
    event = events[0]

    assert isinstance(event, ChangedEvent)
    assert event.id == "foo"
    assert event.rev == doc.rev

    doc2 = await event.doc()
    assert doc2.id == "foo"
    assert doc2.rev == event.rev
    assert "Zebras" in doc2


async def test_deleted_event(filled_database: Database) -> None:
    async def delete_doc(doc: Document) -> None:
        await asyncio.sleep(0.1)

        await doc.delete()

    doc = await filled_database["foo"]

    results = await asyncio.gather(
        listen_for_first_change(filled_database), delete_doc(doc)
    )

    event = results[0]

    assert isinstance(event, DeletedEvent)
    assert event.id == "foo"
    assert event.rev.startswith("2-")
    assert event.sequence is not None


async def test_changed_event_not_include_docs(filled_database: Database) -> None:
    async def update_doc(doc: Document) -> None:
        await asyncio.sleep(0.1)
        doc["Zebras"] = "are black with white stripes"
        await doc.save()

    doc = await filled_database["foo"]

    results = await asyncio.gather(
        listen_for_first_change(filled_database), update_doc(doc)
    )

    event = results[0]

    assert isinstance(event, ChangedEvent)
    assert event.id == "foo"
    assert event.rev == doc.rev
    assert event.sequence is not None

    doc2 = await event.doc()

    assert doc2.id == "foo"
    assert doc2.rev == event.rev
    assert "Zebras" in doc2
    assert doc2["Zebras"] == "are black with white stripes"

    # DO NOT DO THIS, EVIL INTERAL HACK DONE FOR TESTING ONLY
    event.database = cast(Database, None)

    # As we set the database to None, it either gets the Document instance
    # from the local json response, or fails trying to gather it from the remote
    with pytest.raises(AttributeError):
        await event.doc()


async def test_changed_event_include_docs(filled_database: Database) -> None:
    async def update_doc(doc: Document) -> None:
        await asyncio.sleep(0.1)
        doc["Zebras"] = "are black with white stripes"
        await doc.save()

    doc = await filled_database["foo"]

    results = await asyncio.gather(
        listen_for_first_change(filled_database, include_docs=True), update_doc(doc)
    )

    event = results[0]

    assert isinstance(event, ChangedEvent)
    assert event.id == "foo"
    assert event.rev == doc.rev
    assert event.sequence is not None

    assert "doc" in event.json

    doc2 = await event.doc()

    assert doc2.id == "foo"
    assert doc2.rev == event.rev
    assert "Zebras" in doc2
    assert doc2["Zebras"] == "are black with white stripes"

    # DO NOT DO THIS, EVIL INTERAL HACK DONE FOR TESTING ONLY
    event.database = cast(Database, None)

    # As we set the database to None, it either gets the Document instance
    # from the local json response, or fails trying to gather it from the remote
    await event.doc()
